﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace AsyncExamples
{
    public class Worker
    {
        public string Name { get; } //name to recognize our employee
        public Worker(string name) //constructor forcing a name
        {
            Name = name;
        }

        /// <summary>
        /// Ordinary synchronous function
        /// </summary>
        /// <param name="howMuch"></param>
        public void OrdinaryWork(int howMuch)
        {
            Console.WriteLine($"Worker {Name} start ordinary work....");
            for (int i = 1; i <= howMuch; i++)
            {
                Thread.SpinWait(100000000);
                Console.WriteLine($"Worker {Name} has elaborated {i} items");
            }
            Console.WriteLine($"Worker {Name} finished ordinary work.");
        }

        /// <summary>
        /// A method that creates a Task.
        /// </summary>
        /// <param name="howMuch"></param>
        public void WorkTask(int howMuch)
        {
            Console.WriteLine($"Worker {Name} start task working....");
            Task t = new Task(() =>
            {
                for (int i = 1; i <= howMuch; i++)
                {
                    Thread.SpinWait(100000000);
                    Console.WriteLine($"Worker {Name} has elaborated {i} items");
                }
                Console.WriteLine($"Worker {Name} finished task work");
            });
            t.Start();
        }

        /// <summary>
        /// A method that creates and returns a task. This method has become "awaitable"
        /// </summary>
        /// <param name="howMuch"></param>
        /// <returns></returns>
        public Task WorkAsync(int howMuch)
        {
            Console.WriteLine($"Worker {Name} begin async work....");
            Task t = new Task(() =>
            {
                for (int i = 1; i <= howMuch; i++)
                {
                    Thread.SpinWait(100000000);
                    Console.WriteLine($"Worker {Name} has elaborated {i} items");
                }
                Console.WriteLine($"Worker {Name} finished async work");
            });
            t.Start(); //according to the TAP model, the task returned by the asynchronous method should be started
            return t;
        }

        /// <summary>
        /// Async method
        /// </summary>
        /// <param name="howMuch"></param>
        /// <returns></returns>
        public async Task WorkOrderedAsync(int howMuch)
        {
            Console.WriteLine($"Worker {Name} has ordered job....");
            await WorkAsync(howMuch);
            Console.WriteLine($"Worker {Name} received ordered job (used await)");
        }
    }
}
