﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace SimpleTasks
{
    class Program
    {
        static void Main(string[] args)
        {


            #region Start Task
            //As new Task
            Console.WriteLine("\n --------------Casual task Start---------------------");
            Task task = new Task(() =>
            {
                Console.WriteLine("First task is working");
                Thread.Sleep(1000);
                Console.WriteLine("First taks has finished");
            });
            task.Start();
            task.Wait();
            #endregion

            #region Task Run
            //As static method Task.Run
            Console.WriteLine("\n --------------Casual task Run-----------------------");
            task = Task.Run(() =>
            {
                Console.WriteLine("Task Run is working");
                Thread.Sleep(1000);
                Console.WriteLine("Task Run has finished");
            });
            task.Wait();
            #endregion

            #region Task Factory
            //With help of Task.Factory
            Console.WriteLine("\n --------------Task Factory--------------------------");
            task = Task.Factory.StartNew(() =>
            {
                Console.WriteLine("Task Factory is working");
                Thread.Sleep(1000);
                Console.WriteLine("Task Factory has finished");
            });

            task.Wait();
            Console.WriteLine("Press Enter to go to the next example");
            Console.ReadLine();
            #endregion

            #region Many tasks
            //Uruchomienie wielu zadań / starting many tasks
            Console.WriteLine("\n --------------Many tasks---------------------------");
            //An action is a delegate to a method that takes no arguments and returns no value.
            Action a = () =>
            {
                Console.WriteLine($"Task no {Task.CurrentId} has started"); //CurrentId is assigned only at the moment of referring to the task, not at startup
                Thread.SpinWait(new Random().Next(300000000));
                Console.WriteLine($"Task no {Task.CurrentId} has finished");
            };

            List<Task> listOfTasks = new List<Task>();
            for (int i = 0; i < 10; i++)
            {
                listOfTasks.Add(new Task(a));
            }
            listOfTasks.ForEach(t => t.Start());
            listOfTasks.ForEach(t => t.Wait());
            Console.WriteLine("Press Enter to go to the next example");
            Console.ReadLine();
            #endregion

            #region Task with parameter
            //Passing data to the task
            Console.WriteLine("\n --------------Passing data to the task-----------------");
            Task printSthTask = new Task((o) =>
            {
                for (int i = 0; i < 5; i++)
                {
                    Thread.Sleep(200);
                    Console.WriteLine((string)o + " " + i);
                }
            }, "Bla ble blu...");
          
            printSthTask.Start();
            printSthTask.Wait();
            Console.WriteLine("Press Enter to go to the next example");
            Console.ReadLine();
            #endregion

            #region Return Value
            Console.WriteLine("\n --------------Returning value--------------------");
            Task<int> intTask = Task.Run(() =>
            {
                Thread.Sleep(1000);
                return 5;
            });
            Console.WriteLine("We are waiting for the task to return a value");
            var result = intTask.Result; //we are waiting for the end of the task. Only then can we read the result
            Console.WriteLine($"We got value {result}");
            Console.WriteLine("Press Enter to go to the next example");
            Console.ReadLine();
            #endregion

            #region Wait All
            Console.WriteLine("\n --------------Wait for all----------------");
            var tasks = new Task[3];
            for (var i = 0; i < 3; i++)
            {
                tasks[i] = (Task.Run(() =>
                {
                    Console.WriteLine("One of 3 is waiting");
                    Thread.Sleep(i * 500);
                    Console.WriteLine("One of 3 has finished");
                }));
            }
            Console.WriteLine("Waiting for all 3");
            Task.WaitAll(tasks);
            Console.WriteLine("Got all 3");

            var resultTasks = Task.WhenAll(tasks); //Returns the task with everyone
                                                   //  var resultTasks = Task.WhenAll(tasks).Result; //for that to be the case, they have to return something 
            Console.WriteLine("Press Enter to go to the next example");
            Console.ReadLine();
            #endregion

            #region Wrong passing data to the task
            Console.WriteLine("\n --------------Wrong passing data to the task-----");

            var tasks2 = new Task<int>[3];
            //example of wrong use of the variable "i"
            for (var i = 0; i < 3; i++)
            {
                tasks2[i] = (Task<int>.Run(() =>
                {
                    Thread.Sleep(i * 500);
                    return i * 2;//Note! The variable shouldn't be uesed like that
                }));
            }

            var resultTasks2 = Task.WhenAll(tasks2);
            foreach (var item in resultTasks2.Result)
            {
                Console.WriteLine($"The task counted and unfortunately returned: {item}"); //we get the same value because we use the variable "i" which is 3 at the end
            }
            Console.WriteLine("Press Enter to go to the next example");
            Console.ReadLine();
            #endregion

            #region Correct data passing to the task WhenAll
            Console.WriteLine("\n --------------Correct data passing to the task WhenAll---");
            //The version with "i" as parameter will work better
            var tasks3 = new Task<int>[3];
            for (var i = 0; i < 3; i++)
            {
                tasks3[i] = (new Task<int>((o) =>
                {
                    Thread.Sleep((int)o * 500);
                    return (int)o * 2;
                }, i));
            }
            tasks3.ToList<Task>().ForEach((t) => t.Start());
            var resultTasks3 = Task.WhenAll(tasks3);
            resultTasks3.Result.ToList().ForEach((i) => Console.WriteLine($"The task counted and returned correctly: {i}")); //more compact loop
            Console.WriteLine("Press Enter to go to the next example");
            Console.ReadLine();
            #endregion



            #region States of tasks
            //Task states. The example shows all task states
            // Created - created but not started
            // WaitForActivation - It was scheduled to be launched by one of the Continue variants. It is waiting for the end of his predecessor.
            // WaitingToRun - Scheduled to be activated, for example, by Task.Start or TaskFactory.StartNew, but not running. Waits for the TaskScheduler scheduler.
            // Running - task is running and not completed
            // RunToCompletion - The task is complete without interrupting or raising an exception
            // Cancelled - interrupted when the exception was thrown OperationCanceledException (not when Cancel is called on the token)
            // Faulted - interrupted by an exception
            Console.WriteLine("\n --------------States of Tasks----------------------");
            //tasks with child task
            Task normalTask = new Task(() =>
            {
                Console.WriteLine("normalTask has started");
                Thread.Sleep(200);
                Task childTask = Task.Factory.StartNew(() =>
                {
                    Console.WriteLine("childTask  has started");
                    Thread.Sleep(200);
                    Console.WriteLine("childTask  has finished.");
                }, TaskCreationOptions.AttachedToParent); //you can remove this option and see the difference

                Console.WriteLine("normalTask has finished");
            });
            //a task that continues after a normal Task
            Task contiTask = normalTask.ContinueWith((x) =>
            {
                Console.WriteLine("contitask  has started");
                Thread.Sleep(200);
                throw new Exception("OMG We have an Error from contiTask!!!");
              //  Console.WriteLine("contitask has finished");
            }); //you can add a continuation option which does not allow continuation, then we will get caceled status
            CancellationTokenSource cts = new CancellationTokenSource();
            CancellationToken ct = cts.Token;
            
            ///task waiting for cancellation
            Task cancelTask = Task.Factory.StartNew(() =>
            {
                ct.ThrowIfCancellationRequested();
                Console.WriteLine("cancelTask has started");
                Thread.Sleep(500);
                ct.ThrowIfCancellationRequested();
                Console.WriteLine("cancelTask has finished");
            }, ct);
            Console.WriteLine("###     cancelTask Task status just after creation " + cancelTask.Status.ToString()); //In this only case, we will spare the observer because he might not catch such a short state.

            //observer's task. Every 100ms checks the status of all tasks
            Task obserwator = Task.Factory.StartNew(() =>
            {
                int i = 8;
                while (i-- > 0)
                {
                    Thread.Sleep(100);
                    Console.WriteLine("###     normal Task status " + normalTask.Status.ToString());
                    Console.WriteLine("###      conti Task status " + contiTask.Status.ToString());
                    Console.WriteLine("### cancelTask Task status " + cancelTask.Status.ToString());
                }
            });
            
            Thread.Sleep(200);
            normalTask.Start();
            Thread.Sleep(200);
            cts.Cancel();
            try
            {
                Task.WaitAll(normalTask, obserwator, contiTask);
            }
            catch (Exception ex)
            {
                Console.WriteLine($"{ex.InnerException.Message}");
            }
            #endregion
            Console.ReadLine();
        }
    }
}
